/*
 *  Read test cases and exercise localtime() et al.
 *
 *  There are two modes of operation: scripted and "exhaustive".
 *
 *  The default mode is scripted.  Test cases are read from
 *  standard input or named file(s), and executed.  The format of
 *  the test cases is described below.
 *
 *  In "exhaustive" mode, tests are systematically performed over
 *  a selected range of dates.
 *
 *  Options:
 *
 *	-f YYYY-mm-dd	In "exhaustive" mode, start from the given
 *			date (default 1901-12-14)
 *
 *	-i nnn		In "exhaustive" mode, increment by nnn
 *			seconds (default 997)
 *
 *	-s		Call tzset() before running tests.
 *
 *	-t YYYY-mm-dd	In "exhaustive" mode, run to the given
 *			date (default 2038-01-19)
 *
 *	-x		Run in "exhaustive" mode
 *
 *  In "exhaustive" mode, unless adjusted by the -f, -i, or -t
 *  arguments, tests are run on times from 1901-12-14 to 2038-01-19
 *  in increments of 997 seconds (that is, 1901-12-14 00:00:00,
 *  1901-12-14 00:16:37, 1901-12-14 00:33:14, etc.).  At each
 *  timepoint, gmtime, timegm, mktime, and localtime are tested.
 *  mktime and localtime are tested using the system's default
 *  time zone, e.g. as set by the TZ variable.
 *
 *  Test Scripts
 *
 *  A test script is a text file.  Each line is interpreted as
 *  whitespace-separated tokens.  Any text following a '#' character
 *  is a comment to be ignored.
 *
 *  The first token on each line is a keyword performing one
 *  action.  Remaining tokens are arguments.
 *
 *  The keywords are as follows:
 *
 *	gmtime t date time [options]
 *
 *	Applying gmtime() to the time_t value t is expected to
 *	give the indicated data and time.  Contrariwise, applying
 *	timegm() to the date and time is expected to give the
 *	indicated time_t value.
 *
 *	The options are described below.
 *
 *	Example:
 *		gmtime  420206400  1983-04-26 12:00:00  yd=115 wd=2
 *
 *
 *	localtime t date time [options]
 *
 *	Applying localtime() to the time_t value t, under the
 *	current time zone, is expected to give the indicated data
 *	and time.  Contrariwise, applying mktime() to the date
 *	and time is expected to give the  indicated time_t value.
 *
 *	The options are described below.
 *
 *	Example:
 *		localtime  482432400  1985-04-15 12:00:00  z=EST o=-18000
 *
 *	tz zonename
 *
 *	Set the time zone to be used for following localtime tests.
 *
 *
 *  Options
 *
 *  For gmtime and localtime tests, a number of optional
 *  parameters may be specified.  These options are specified using
 *  a simple keyword=value in columns 5 and beyond.  Whitespace is
 *  not permitted on either side of the = sign.  Unrecognized
 *  options are quietly ignored.
 *
 *  The available options are:
 *
 *	wd=n	Specify the expected tm_wday value for localtime
 *		and gmtime conversion (0 = Sunday, 6 = Saturday).
 *
 *	yd=n	Specify the expected tm_yday value for localtime
 *		and gmtime conversion.
 *
 *	dst=n	Specify the expected tm_isdst value for localtime
 *		conversion.
 *
 *	z=name	Specify the expected tm_zone value for localtime
 *		conversion (if tm_zone is supported).
 *
 *	o=nnn	Specify the expected tm_gmtoff value for localtime
 *		conversion (if tm_gmtoff is supported).
 *
 *	fl=fff	Specify optional processing flags.  fff is a
 *		comma-separated list of flag names.  The following
 *		flags are supported:
 *
 *			noi	Do not attempt to perform the
 *				inverse mktime or timegm test.
 *				(This flag is typically used for
 *				localtime tests during DST "fall
 *				back" intervals, when the time is
 *				ambiguous and does not uniquely
 *				correspond to a single time_t value.)
 *
 *			denorm	The broken-down struct tm is
 *				denormalized and would never be
 *				generated by localtime or gmtime;
 *				perform *only* the inverse mktime
 *				or timegm test.
 *
 *	See the sample testcases file for examples of all these
 *	options and flags in use.
 *
 *  Known Issues
 *
 *  The zone names used in a test script must (obviously) match
 *  those installed along with localtime et al. on the system under
 *  test.  Beware that different OSes and distributions have used
 *  significantly different sets of names ("America/New York",
 *  "US/Eastern", "EST5EDT", etc.)
 *
 *  Different operating systems use different mechanisms to set
 *  the default time zone for the current machine.  When driven by
 *  a test script, all tests use zones as explicitly specified in
 *  the script.  But exhaustive (-x) tests rely on the system's default
 *  time zone.  If the system's notion of its default time zone is
 *  stored in a way that the copy of localtime under test is not aware
 *  of, it may be necessary to explicitly set a zone, e.g. by setting
 *  the environment variable TZ.  (It may also help to use the -s
 *  option, causing an explicit call to tzset().)
 *
 *  At the moment, there are tests on localtime, gmtime, mktime, and timegm.
 *  Tests on ctime and strftime would be useful but are not implemented yet.
 *  There are also no tests on asctime, nor on tzset, tzsetwall, offtime,
 *  or timeoff.
 *
 *  Written by Steve Summit, 2015-12.
 *  This code is in the Public Domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <setjmp.h>
#include <errno.h>

/* temporarily hardwired (could get these from private.h) */
#define TM_ZONE tm_zone
#define TM_GMTOFF tm_gmtoff

#define EPOCH_WDAY 4	/* Thursday */

#define Streq(s1, s2) (strcmp(s1, s2) == 0)

int getargs(char **, char *, int);

void runtests(FILE *, const char *);
void run_exhaustive(const struct tm *, const struct tm *, time_t);

time_t quick_timegm(const struct tm *);
int isleap(int);
int nleapsbetween(int, int);

int tmcmp(const struct tm *, const struct tm *);

void catchsig(int);

/* our own copy of this time-honored array (would be nice to share) */
int monthlens[2][12] = {
	{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
	{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
};

int errs = 0;

jmp_buf jmpbuf;

const char *progname = "testsuite";

int main(int argc, char *argv[])
{
	int ferrs = 0;
	int exhaustive = 0;
	int argi = 1;
	int calltzset = 0;
	char *fromstr = "1901-12-14";
	char *tostr = "2038-01-19";
	char *incstr = "997";	/* rather arbitrary, but useful if prime */
	struct tm tm;

	for(; argi < argc && argv[argi][0] == '-'; argi++) {
		char *p;
		for(p = &argv[argi][1]; *p != '\0'; p++) {
			switch(*p) {
				case 'f':
					fromstr = argv[++argi];
					break;

				case 'i':
					incstr = argv[++argi];
					break;

				case 's':
					calltzset = 1;
					break;

				case 't':
					tostr = argv[++argi];
					break;

				case 'x':
					exhaustive = 1;
					break;

				default:
					fprintf(stderr,
						"%s: unknown option -%c\n",
								progname, *p);
			}
		}
	}

	signal(SIGSEGV, catchsig);
	signal(SIGALRM, catchsig);

	/* TODO: check for -x plus input file, other mismatched args */

	if(calltzset) {
		tzset();
	}

	if(exhaustive) {
		struct tm from, to;
		time_t inc;

		if(sscanf(fromstr, "%d-%d-%d",
			    &from.tm_year, &from.tm_mon, &from.tm_mday) != 3) {
			fprintf(stderr,
				    "%s: syntax error in from string \"%s\"\n",
							progname, fromstr);
			}
		from.tm_year -= 1900;
		from.tm_mon--;
		from.tm_hour = from.tm_min = from.tm_sec = 0;
		from.tm_isdst = -1;
		from.tm_wday = 0;
		from.tm_yday = 0;

		if(sscanf(tostr, "%d-%d-%d",
				  &to.tm_year, &to.tm_mon, &to.tm_mday) != 3) {
			fprintf(stderr,
				    "%s: syntax error in to string \"%s\"\n",
							progname, tostr);
		}
		to.tm_year -= 1900;
		to.tm_mon--;
		to.tm_hour = to.tm_min = to.tm_sec = 0;
		to.tm_isdst = -1;
		to.tm_wday = 0;
		to.tm_yday = 0;

		inc = atol(incstr);

		run_exhaustive(&from, &to, inc);
	} else if(argi == argc) {
		printf("%s: reading tests from standard input\n", progname);
		runtests(stdin, "standard input");
	} else {
		FILE *fp;
		for(; argi < argc; argi++) {
			fp = fopen(argv[argi], "r");
			if(fp == NULL) {
				fprintf(stderr, "%s: can't open %s: %s\n",
					progname, argv[argi], strerror(errno));
				ferrs++;
				continue;
			}
			runtests(fp, argv[argi]);
			fclose(fp);
		}
	}

	if(errs != 0 || ferrs != 0) {
		printf("%d tests failed", errs);
		if(ferrs != 0)
			printf(", %d other errors", ferrs);
		printf("\n");
	}

	if(ferrs != 0) return 2;
	if(errs != 0) return 1;
	return 0;
}

#define LOCAL 01
#define NOINV 02
#define DENORM 04

void basictest(time_t t, const struct tm *itmp, int flags, const char *tag)
{
	int sig;
	time_t t2;
	struct tm tm, tm2;
	struct tm *tmp;

	tm = *itmp;

	/* Ye olde C-style exception handling */
	if((sig = setjmp(jmpbuf)) != 0) {
		printf("%s: %s\n", tag, sig == SIGALRM ? "taking too long" :
						         "exception");
		errs++;
		return;	/* skips further tests for this test case */
	}

	alarm(1);

	if(!(flags & DENORM)) {

		if(flags & LOCAL) {
			tmp = localtime(&t);
		} else {
			tmp = gmtime(&t);
		}

		if(tmp == NULL) {
			printf("%s: %s failed\n", tag,
			     (flags & LOCAL) ? "localtime" : "gmtime");
			errs++;
			return;	/* skips further tests for this test case */
		}

		tm2 = *tmp;

		if(tmcmp(&tm, &tm2) != 0) {
			printf("%s: expected %04d-%02d-%02d %02d:%02d:%02d, got %04d-%02d-%02d %02d:%02d:%02d\n",
				tag,
				tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday,
					tm.tm_hour, tm.tm_min, tm.tm_sec,
				tm2.tm_year+1900, tm2.tm_mon+1, tm2.tm_mday,
					tm2.tm_hour, tm2.tm_min, tm2.tm_sec);
			errs++;
		}
		if(tm.tm_wday != -1) {
			if(tm2.tm_wday != tm.tm_wday) {
				printf("%s: expected wday %d, got %d\n",
					tag, tm.tm_wday, tm2.tm_wday);
				errs++;
			}
		}
		if(tm.tm_yday != -1) {
			if(tm2.tm_yday != tm.tm_yday) {
				printf("%s: expected yday %d, got %d\n",
					tag, tm.tm_yday, tm2.tm_yday);
				errs++;
			}
		}
		if(tm.tm_isdst != -1) {
			if(tm2.tm_isdst != tm.tm_isdst) {
				printf("%s: expected dst %d, got %d\n",
					tag, tm.tm_isdst, tm2.tm_isdst);
				errs++;
			}
		}
#ifdef TM_ZONE
		if(tm.TM_ZONE && *tm.TM_ZONE) {
			if(!Streq(tm2.TM_ZONE, tm.TM_ZONE)) {
				printf("%s: expected \"%s\", got \"%s\"\n",
					tag, tm.TM_ZONE, tm2.TM_ZONE);
				errs++;
			}
		}
#endif
#ifdef TM_GMTOFF
		if(tm.TM_GMTOFF != -1) {
			if(tm2.TM_GMTOFF != tm.TM_GMTOFF) {
				printf("%s: expected offset %ld, got %ld\n",
					tag, tm.TM_GMTOFF, tm2.TM_GMTOFF);
				errs++;
			}
		}
#endif
	}

	/* now inverse */
	if(!(flags & NOINV)) {

		tm.tm_isdst = -1;
		tm.tm_wday = 0;
		tm.tm_yday = 0;

		if(flags & LOCAL) {
			t2 = mktime(&tm);
		} else {
#ifdef TEST_QUICK_TIMEGM
			t2 = quick_timegm(&tm);
#else
			t2 = timegm(&tm);
#endif
		}

		if(t2 != t) {
			printf("%s: inverse expected %ld, got %ld\n",
					tag, t, t2);
			errs++;
		}
	}

	alarm(0);
}

#define MAXLINE 512
#define MAXARGS 10

void runtests(FILE *fp, const char *name)
{
	int lineno = 0;
	char line[MAXLINE];
	char *p;
	char *av[MAXARGS];
	int ac;
	int sig;
	time_t t, t2;
	struct tm tm, tm2;
	struct tm *tmp;
	char tag[100];
	int i;
	int flags;

	while(fgets(line, sizeof(line), fp) != NULL) {
		lineno++;
		if((p = strrchr(line, '\n')) != NULL) *p = '\0';
		if((p = strrchr(line, '\r')) != NULL) *p = '\0';
		if((p = strchr(line, '#')) != NULL) *p = '\0';
		if(*line == '\0') continue;
		ac = getargs(av, line, MAXARGS);
		if(ac == 0) continue;

		snprintf(tag, sizeof(tag), "%s, line %d: %s",
							name, lineno, av[0]);

		if(Streq(av[0], "tz")) {
			setenv("TZ", av[1], 1);
		} else if(Streq(av[0], "localtime") || Streq(av[0], "gmtime")) {
			if(ac < 4) {
				fprintf(stderr,
				    "%s: %s, line %d: %s: too few arguments\n",
					progname, name, lineno, av[0]);
				continue;
			}

			t = atoll(av[1]);

			if(sscanf(av[2], "%d-%d-%d",
			    &tm.tm_year, &tm.tm_mon, &tm.tm_mday) != 3 ||
				sscanf(av[3], "%d:%d:%d",
				   &tm.tm_hour, &tm.tm_min, &tm.tm_sec) != 3) {
				fprintf(stderr,
				 "%s: %s, line %d: %s: bad date/time syntax\n",
					progname, name, lineno, av[0]);
				continue;
			}

			tm.tm_year -= 1900;
			tm.tm_mon--;

			flags = 0;
			tm.tm_wday = -1;
			tm.tm_yday = -1;
			tm.tm_isdst = -1;
#ifdef TM_ZONE
			tm.TM_ZONE = NULL;
#endif
#ifdef TM_GMTOFF
			tm.TM_GMTOFF = -1;
#endif
			if(Streq(av[0], "localtime")) flags |= LOCAL;

			for(i = 4; i < ac; i++) {
				char *k = av[i];
				char *v = strchr(av[i], '=');
				if(v == NULL) {
					fprintf(stderr,
			    "%s: %s, line %d: %s bad aux parm syntax \"%s\"\n",
						progname, name, lineno,
							av[0], av[i]);
					continue;
				}
				*v++ = '\0';
				if(Streq(k, "fl")) {
					/* scan comma-separated list */
					for(p = v; p && *p; ) {
						char *p2 = strchr(p, ',');
						if(p2 != NULL) *p2 = '\0';
						if(Streq(p, "noi"))
							flags |= NOINV;
						else if(Streq(p, "denorm"))
							flags |= DENORM;
						if(p2) p = p2 + 1;
						else   p = NULL;
					}
				} else if(Streq(k, "wd")) {
					tm.tm_wday = atoi(v);
				} else if(Streq(k, "yd")) {
					tm.tm_yday = atoi(v);
				} else if(Streq(k, "dst")) {
					tm.tm_isdst = atoi(v);
#ifdef TM_ZONE
				} else if(Streq(k, "z")) {
					tm.TM_ZONE = v;
#endif
#ifdef TM_GMTOFF
				} else if(Streq(k, "o")) {
					tm.TM_GMTOFF = atol(v);
#endif
				}
			}

			basictest(t, &tm, flags, tag);

		} else {
			fprintf(stderr,
		      "%s: %s, line %d: warning: unrecognized operator '%s'\n",
				progname, name, lineno, av[0]);
		}
	}
}

void run_exhaustive(const struct tm *from, const struct tm *to, time_t inc)
{
	time_t t, t2;
	struct tm tm, *tmp, tmcopy;
	long ntests = 0;
	char tag[30];
	time_t tmpx, tmpi;
	int i;
	static int beenhere = 0;

	printf("exhaustive tests: from %d-%02d-%02d ",
			from->tm_year+1900, from->tm_mon+1, from->tm_mday);
	printf("to %d-%02d-%02d by %ld\n",
		to->tm_year+1900, to->tm_mon+1, to->tm_mday, inc);

	tm = *from;
	t = quick_timegm(from);

	tm.tm_wday = (t / 86400 + EPOCH_WDAY) % 7;
	if(tm.tm_wday < 0) tm.tm_wday += 7;

	tm.tm_yday = 0;
	for(i = 0; i < tm.tm_mon; i++)
		tm.tm_yday += monthlens[isleap(1900+tm.tm_year)][i];
	tm.tm_yday += tm.tm_mday - 1;

	while(tmcmp(&tm, to) <= 0) {

		ntests++;

		snprintf(tag, sizeof(tag), "%ld", t);

		/*
		 *  This do/while loop runs exactly once, and is here just
		 *  so that a break will skip the rest of the current test
		 *  and jump down to the increment to the next one.
		 *  (These breaks are therefore even more like gotos
		 *  than usual.)
		 */
		do {
			tm.tm_isdst = -1;
#ifdef TM_ZONE
			tm.TM_ZONE = NULL;
#endif
#ifdef TM_GMTOFF
			tm.TM_GMTOFF = 0;
#endif
			/*
			 *  The strategy for exhaustive tests is a little
			 *  different than for the explicit, script-driven
			 *  ones.  Hor gmtime, we can reasonably precompute
			 *  a (time, struct tm) pair, so we can call
			 *  basictest() more or less as for the scripted
			 *  case.  For localtime, though, since we can't
			 *  reasonably predict, and since we therefore have
			 *  little choice but to do a more simpleminded
			 *  end-to-end test, let's run it the other way,
			 *  starting with the struct tm which the GMT-based
			 *  test computed, running it struct tm -> time ->
			 *  struct tm, and then checking to see whether the
			 *  tm_gmtoff makes sense given the (GMT) time we
			 *  originally started with.
			 */

			basictest(t, &tm, 0, tag);

			tmcopy = tm;

			t2 = mktime(&tmcopy);

			if(t2 == -1) {
				if(tm.tm_hour == 2) {
					/* assume ambiguous DST (maybe temporary) */
					break;	/* continue to next case */
				}

				printf("%s: %s failed\n", tag, "mktime");
				errs++;
				break;	/* skips further tests for this case */
			}

			tmp = localtime(&t2);
			if(tmp == NULL) {
				printf("%s: %s failed\n", tag, "localtime");
				errs++;
				break;	/* skips further tests for this case */
			}

			if(tmcmp(&tm, tmp) != 0) {
				printf("%s: expected %04d-%02d-%02d %02d:%02d:%02d, got %04d-%02d-%02d %02d:%02d:%02d\n",
				    tag,
				    tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday,
					tm.tm_hour, tm.tm_min, tm.tm_sec,
				    tmp->tm_year+1900, tmp->tm_mon+1, tmp->tm_mday,
					tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
				errs++;
			}

#ifdef TM_GMTOFF
			if(tmp->TM_GMTOFF == 0 && !beenhere) {
				fprintf(stderr,
	"%s: no or GMT/UTC time zone set, so no nontrivial tz offset tested\n",
								progname);
				fprintf(stderr,
				       "try setting TZ or invoking with -s\n");
				beenhere = 1;
			}

			if(t - t2 != tmp->TM_GMTOFF) {
				printf(
				"%s: tm_gmtoff not as expected: %ld vs. %ld\n",
						tag, t - t2, tmp->TM_GMTOFF);
				errs++;
			}
#endif
		} while(0);

		t += inc;

		/* have to worry about overflow, since time_t might be */
		/* 64 bits, while tm_sec et al. almost certainly aren't */

		tmpx = tm.tm_sec + inc;
		tm.tm_sec = tmpx % 60;
		tmpi = tmpx / 60;

		tmpx = tm.tm_min + tmpi;
		tm.tm_min = tmpx % 60;
		tmpi = tmpx / 60;

		tmpx = tm.tm_hour + tmpi;
		tm.tm_hour = tmpx % 24;
		tmpi = tmpx / 24;

		tm.tm_mday += tmpi;	/* assuming no more ovrfl possib */
		tm.tm_wday = (tm.tm_wday + tmpi) % 7;
		tm.tm_yday += tmpi;	/* overflow handled in mday loop */

		while(tm.tm_mday > monthlens[isleap(1900+tm.tm_year)][tm.tm_mon]) {
			tm.tm_mday -= monthlens[isleap(1900+tm.tm_year)]
					       [tm.tm_mon];
			tm.tm_mon++;

			if(tm.tm_mon >= 12) {
				tm.tm_yday -= isleap(1900+tm.tm_year) ? 366
								      : 365;
				tm.tm_year++;
				tm.tm_mon -= 12;
			}
		}
	}

	printf("%ld tests completed\n", ntests);
}

int tmcmp(const struct tm *tm1, const struct tm *tm2)
{
	if(tm1->tm_year != tm2->tm_year)
		return (tm1->tm_year < tm2->tm_year) ? -1 : 1;
	else if(tm1->tm_mon != tm2->tm_mon)
		return (tm1->tm_mon < tm2->tm_mon) ? -1 : 1;
	else if(tm1->tm_mday != tm2->tm_mday)
		return (tm1->tm_mday < tm2->tm_mday) ? -1 : 1;
	else if(tm1->tm_hour != tm2->tm_hour)
		return (tm1->tm_hour < tm2->tm_hour) ? -1 : 1;
	else if(tm1->tm_min != tm2->tm_min)
		return (tm1->tm_min < tm2->tm_min) ? -1 : 1;
	else if(tm1->tm_sec != tm2->tm_sec)
		return (tm1->tm_sec < tm2->tm_sec) ? -1 : 1;
	else return 0;
}

time_t quick_timegm(const struct tm *tmp)
{
	int m;
	time_t r = (time_t)(1900 + tmp->tm_year - 1970) * 365 * 86400;
	r += nleapsbetween(1970, 1900 + tmp->tm_year-1) * (time_t)86400;
#ifdef DEBUG
	printf("quick_timegm: nleapsbetween: %d\n",
				nleapsbetween(1970, 1900 + tmp->tm_year-1));
#endif
	for(m = 0; m < tmp->tm_mon; m++)
		r += monthlens[isleap(1900+tmp->tm_year)][m] * 86400;
	r += (tmp->tm_mday - 1) * 86400;
	r += (tmp->tm_hour * 60 + tmp->tm_min) * 60 + tmp->tm_sec;
#ifdef DEBUG
	printf("quick_timegm: %04d-%02d-%02d %02d:%02d:%02d -> %ld\n",
		1900+tmp->tm_year, tmp->tm_mon+1, tmp->tm_mday,
			tmp->tm_hour, tmp->tm_min, tmp->tm_sec, r);
#endif
	return r;
}

int isleap(int y)
{
	return y % 4 == 0 && (y % 100 != 0 || y % 400 == 0);
}

int nleapsbetween(y1, y2)
{
	if(y1 == y2) {
		return 0;	/* unnecessary optimization */
	} else if(y1 != 0) {
		return nleapsbetween(0, y2) - nleapsbetween(0, y1);
	} else {
		/*
		 *  This could simply be
		 *	return y2 / 4 - y2 / 100 + y2 / 400;
		 *  if only C guaranteed Euclidean division...
		 */
		int ret = 0;
		int q = y2 / 4;
		if(y2 % 4 < 0) q--;
		ret += q;
		q = y2 / 100; if(y2 % 100 < 0) q--;
		ret -= q;
		q = y2 / 400; if(y2 % 400 < 0) q--;
		ret += q;
		return ret;
	}
}

void catchsig(int sig)
{
	longjmp(jmpbuf, sig);
}

/*
 *  takes a string (line) and builds an array of pointers to each word in it.
 *  words are separated by spaces or any control characters.  At most maxargs
 *  pointers are calculated.  \0's are inserted in line, so that each word
 *  becomes a string in its own right.  The number of pointers (argc) is
 *  returned.
 */

#include <stdio.h>

#define iswhite(c) ((c) == ' ' || (c) == '\t' || (c) == '\n')

getargs(char *argv[], char *line, int maxargs)
{
	int nargs;

	nargs = 0;

	while(1) {
		while(iswhite(*line))
			line++;

		if(*line == '\0') {
			if(nargs < maxargs) *argv = NULL;
			return(nargs);
		}

		*argv++ = line;
		nargs++;

		while(!iswhite(*line) && *line != '\0')
			line++;

		if(*line == '\0') {
			if(nargs < maxargs) *argv = NULL;
			return(nargs);
		}
		*line++ = '\0';
		if(nargs == maxargs) return(nargs);
	}
}
